-- Written by Team Ulysses, http://ulyssesmod.net/
AddCSLuaFile( "autorun/sh_uclip.lua" )
AddCSLuaFile( "autorun/cl_uclip.lua" )

module( "Uclip", package.seeall )

------------
-- Config --
------------

function adminCheck( ply ) -- Who can move through everything? Comment out the line below if you don't want anyone moving through anything but their own stuff.
	--if true then return false end
	if ulx and ply:query( "ulx noclip" ) then 
		return true
	elseif ply:IsAdmin() or ply:IsSuperAdmin() then 
		return true 
	end
	
	return false
end

maxrecurse = 4 -- Max recurses. Recursion is used when we need to test a new velocity.
-- Should rarely recurse more than twice or else objects are probably wedged together. 
-- (We have checks for wedged objects too so don't worry about it). 0 = disable (not recommended but shouldn't have any problems)

maxloop = 50 -- Max loops. We need to loop to find objects behind other objects. 0 = infinite (not recommended but shouldn't have any problems)
-- This *could* open an exploit, but users would be hard pressed to use it... so we'll see.

--------------------------------------------------------
-- End config. DO NOT MODIFY ANYTHING BELOW THIS LINE!--
--------------------------------------------------------

noProtection = false -- If there's no protection whatsoever, this is flagged. 
-- We need this flag because with a protector, we default to _not_ being able to go through things.
-- This flag saves us major memory/bandwidth when there's no protection
	
function checkOwnership( ply, ent )
	if ent:IsPlayer() or ent:IsNPC() then return true end -- Let players go through other players and npcs (too problematic otherwise)
	if noProtection then return true end -- No protection, they own everything.
	
	updateOwnership( ply, ent )	-- Make sure server and the client are current.
	return ent.Uclip[ ply ] -- This is flagged by updateOwnership. True if they own it, nil otherwise.
end

-- Given a velocity, normalized velocity, and a normal:
-- Calculate the velocity toward the wall and remove it so we can "slide" across the wall.
function calcWallSlide( vel, normal )
	local toWall = normal * -1
	local velToWall = vel:Dot( toWall ) * toWall
	return vel - velToWall
end

function zeromove( move )
	move:SetVelocity( Vector( 0, 0, 0 ) )
	move:SetVelocity( Vector( 0, 0, 0 ) )
	move:SetForwardSpeed( 0 )
	move:SetSideSpeed( 0 )	
	move:SetUpSpeed( 0 )
end

-- Check if a player is stuck in an object or in the world.
-- Returns true and ent they're stuck on if they're stuck.
function isStuck( ply, filter )
	local ang = ply:EyeAngles()
	local directions = {
		ang:Right(),
		ang:Right() * -1,
		ang:Forward(),		
		ang:Forward() * -1,
		ang:Up(),
		ang:Up() * -1,
	}
	local ents = {}
	
	local t = {}
	t.start = ply:GetPos()
	t.filter = filter
	
	-- Check if they're stuck by checking each direction. A minimum number of directions should hit the same object if they're stuck.
	for _, dir in ipairs( directions ) do
		t.endpos = ply:GetPos() + dir
		local tr = util.TraceEntity( t, ply )
		if tr.Entity:IsValid() and tr.HitPos == tr.StartPos then
			ents[ tr.Entity ] = ents[ tr.Entity ] or 0
			ents[ tr.Entity ] = ents[ tr.Entity ] + 1
		end
	end
	
	for ent, hits in pairs( ents ) do
		if hits >= 4 then
			return true, ent
		end
	end
	
	return false
end

-- The brain of Uclip, this makes sure they can move where they want to.
function checkVel( ply, move, vel, recurse, hitnorms )
	if vel == Vector( 0, 0, 0 ) then return end -- No velocity, don't bother.
	
	local ft = FrameTime()
	local veln = vel:GetNormalized()
	hitnorms = hitnorms or {} -- This is used so we won't process the same normal more than once. (IE, we don't get a wedge where we have to process velocity to 0)
	
	recurse = recurse or 0 -- Keep track of how many recurses
	recurse = recurse + 1
	if recurse > maxrecurse and maxrecurse > 0 then -- Hard break
		zeromove( move )
		return
	end
	
	local t = {}
	t.start = ply:GetPos()
	t.endpos = ply:GetPos() + vel * ft + veln -- Add an extra unit in the direction they're headed just to be safe.
	t.filter = { ply }
	local tr = util.TraceEntity( t, ply )

	local loops = 0
	while tr.Hit do -- Recursively check all the hits. This is so we don't miss objects behind another object.
		loops = loops + 1
		if maxloop > 0 and loops > maxloop then 
			zeromove( move )			
			return
		end
		
		if tr.HitWorld or ( tr.Entity:IsValid() and (tr.Entity:GetClass() == "prop_dynamic" or (not checkOwnership( ply, tr.Entity ) and not isStuck( ply, t.filter ))) ) then -- If world or a prop they don't own that they're not stuck inside. Ignore prop_dynamic due to crash.
			local slide = calcWallSlide( vel, tr.HitNormal )
			move:SetVelocity( slide )			
			move:SetSideSpeed( 0 )
			move:SetForwardSpeed( 0 )
			move:SetUpSpeed( 0 )
			
			if table.HasValue( hitnorms, tr.HitNormal ) then -- We've already processed this normal. We can get this case when the player's noclipping into a wedge.
				zeromove( move )			
				return
			end
			table.insert( hitnorms, tr.HitNormal )					
			
			return checkVel( ply, move, slide, recurse, hitnorms ) -- Return now so this func isn't left on stack			
		end

		if tr.Entity and tr.Entity:IsValid() then -- Ent to add!
			table.insert( t.filter, tr.Entity )		
		end

		tr = util.TraceEntity( t, ply )
	end
end

function move( ply, move )
	if ply:GetMoveType() ~= MOVETYPE_NOCLIP then return end -- Not interested in anything but noclip
	if adminCheck( ply ) then return end -- They can go through everything
	if SERVER and server_settings.Int( "sbox_noclip" ) == 2 then return end -- They can go through everything. Note this is only checked on servers so prediction won't work... but it's better than nothing.
	local ft = FrameTime()

	local vel = move:GetVelocity() -- How far are they going to move this frame?
	-- We're tacking some extra velocity on because we don't have how much they're going to accelerate this frame.
	vel = vel + ply:EyeAngles():Forward() * move:GetForwardSpeed() / 4 * ft
	vel = vel + ply:EyeAngles():Right() * move:GetSideSpeed() / 4 * ft
	vel = vel + ply:EyeAngles():Up() * move:GetUpSpeed() / 3 * ft
	-- We're not doing any extra checks we should be here (approaching max, slowing down, etc) because it really doesn't matter enough to us.
	
	checkVel( ply, move, vel )
end
hook.Add( "Move", "UclipMove", move )